@echo off

rem Local variables only
setlocal

rem ****************************************
rem Dependency versions must be defined here
rem ****************************************
set hdf5=hdf5-1.10.9
set med=med-4.1.1
set metis=metis-5.1.0
set msmpi=msmpi-10.1.2
set openblas=openblas-0.3.21
set scalapack=scalapack-2.1.0
set mumps=mumps-5.4.1
set aed2=libaed2-1.2.0
set gotm=gotm-2019-06-14-opentelemac

rem *********************************************************************************
rem Parse input arguments to determine where and how the dependencies should be built
rem *********************************************************************************

rem Check that the installation path was provided
set install_path=%~1
if "%install_path:~0,1%" == "/" set install_path=
if "%install_path%" == "" (
  echo The syntax is:
  echo:
  echo %~n0 installation_path [compiler]
  echo:
  echo Description:
  echo 	installation_path:	Path where TELEMAC-MASCARET dependencies will be installed.
  echo 	compiler (optional^):	Compiler that should be used, either mingw (default^) or intel.
  exit /b 0
)

rem Ensure that the installation path is absolute and create it
set install_path=%~f1
mkdir %install_path% 2>nul
rem Convert directory separators to forward slashes for CMake (they are mandatory because ScaLAPACK uses a bugged CMake module)
set install_path=%install_path:\=/%

rem Set the compiler (default is mingw)
if "%~2" == "" (
  set compiler=mingw
) else (
  set compiler=%~2
)

rem Store the compiler name with the first letter capitalized
for /f "tokens=2 delims=:" %%a in ('"find "" ":%compiler:~,1%" 2>&1"') do set compiler_name=%%a%compiler:~1%

rem ***********************************************************************
rem Check that all prerequisites (Git, Cmake and Python 3.10) are available 
rem ***********************************************************************

rem Check for the presence of Git
set git_path=
for /f "delims=" %%a in ('"where git 2>nul"') do set git_path=%%a
if not defined git_path (
  echo Error: Git not found. Please ensure that Git is in your PATH.
  exit /b 0
)

rem Check for the presence of CMake
set cmake_path=
for /f "delims=,tokens=*" %%a in ('"where cmake 2>nul"') do set cmake_path=%%a
if not defined cmake_path (
  echo Error: CMake not found. Please ensure that CMake is in your PATH.
  exit /b 0
)

rem Check for the presence of the Python launcher
set py_path=
for /f "delims=,tokens=*" %%a in ('"where py 2>nul"') do set py_path=%%a
if not defined py_path (
  echo Error: Python launcher not found. Please install Python 3.10 with the launcher.
  exit /b 0
)
rem Check for the presence of Python 3.10
py -0 2>nul | findstr /I /B /C:" -3.10-64" > nul 2>&1
if %errorlevel% == 1 (
    echo Error: Python 3.10 not found. You need to install it using the official installer.
    exit /b 0
)

rem **********************
rem Build the dependencies
rem **********************

pushd %~dp0deps

if %compiler% == mingw (
  goto mingw_build
)

if %compiler% == intel (
  goto intel_build
)

echo Unsupported compiler: %compiler%

:exit
endlocal
popd
color 7
exit /b 0

:mingw_build
rem Try to find Mingw in the current PATH
set mingw_path=
for /f "delims=" %%a in ('"where mingw32-make 2>nul"') do set mingw_path=%%a
rem If not found, add the usual Mingw location to PATH
if not defined mingw_path set path=C:\mingw64\bin;%path%
for /f "delims=" %%a in ('"where mingw32-make 2>nul"') do set mingw_path=%%a
if not defined mingw_path (
  echo Error: Mingw not found. Please ensure that Mingw bin folder is in your PATH.
  goto exit
)

call :build_hdf5 || goto exit
call :build_med || goto exit
call :build_metis || goto exit
call :build_msmpi || goto exit
call :build_openblas || goto exit
call :build_scalapack || goto exit
call :build_mumps || goto exit
call :build_aed2 || goto exit
call :build_gotm || goto exit
call :install_python_packages || goto exit
goto exit

:intel_build
call :build_hdf5 || goto exit
call :build_med || goto exit
call :build_metis || goto exit
rem call :build_mumps || goto exit
rem call :build_aed2 || goto exit
rem call :buid_gotm || goto exit
call :install_python_packages || goto exit
goto exit

rem ****
rem HDF5
rem ****
:build_hdf5
call :download %hdf5% "https://support.hdfgroup.org/ftp/HDF5/releases/hdf5-1.10/hdf5-1.10.9/src/hdf5-1.10.9.zip" || exit /b 1
set cmake_config_options=-DONLY_SHARED_LIBS:BOOL=ON -DHDF5_GENERATE_HEADERS:BOOL=OFF -DBUILD_TESTING:BOOL=OFF -DHDF5_BUILD_EXAMPLES:BOOL=OFF -DHDF5_BUILD_TOOLS:BOOL=OFF -DHDF5_BUILD_UTILS:BOOL=OFF -DHDF5_BUILD_HL_LIB:BOOL=OFF -DHDF5_BUILD_HL_TOOLS:BOOL=OFF
call :cmake_build %hdf5% || exit /b 1
exit /b 0

rem ***
rem MED
rem ***
:build_med
call :download %med% "https://files.salome-platform.org/Salome/other/med-4.1.1.tar.gz" || exit /b 1
set cmake_config_options=-DHDF5_ROOT_DIR:PATH=%install_path%/%hdf5% -DMEDFILE_BUILD_TESTS:BOOL=OFF -DMEDFILE_INSTALL_DOC:BOOL=OFF
if %compiler% == intel (
  rem MED libraries cannot be built as shared libraries with Visual Studio generator as they
  rem miss the proper __declspec(dllexport) declarations on each class and function
  set cmake_config_options=%cmake_config_options% -DMEDFILE_BUILD_SHARED_LIBS:BOOL=OFF -DMEDFILE_BUILD_STATIC_LIBS:BOOL=ON
)
call :cmake_build %med% || exit /b 1
exit /b 0

rem *****
rem METIS
rem *****
:build_metis
call :download %metis% "https://github.com/scivision/METIS/archive/refs/tags/v5.1.0.3.tar.gz" || exit /b 1
set cmake_config_options=-DBUILD_SHARED_LIBS:BOOL=OFF
call :cmake_build %metis% || exit /b 1
exit /b 0

rem ******
rem MS-MPI
rem ******
:build_msmpi
echo Building %msmpi% with %compiler_name% compiler...
set carch=x86_64
pushd %msmpi%
cd sdk
rd /s/q build-%carch% 2>nul
mkdir build-%carch%
cd build-%carch%
mkdir bin include lib
rem Generate import library
dlltool -k -d ..\msmpi.def.%carch% -l lib\libmsmpi.a || popd && exit /b 1
rem Headers
copy ..\mpi.h include\
copy ..\mpif.h include\
copy ..\mpi.f90 include\
copy ..\mpifptr.h.%carch% include\mpifptr.h
rem Build compiler wrappers
set mpic=..\mpi.c
set cflags=-s -O2 -DNDEBUG
gcc %cflags% -o bin\mpicc.exe -DCC %mpic% || popd && exit /b 1
gcc %cflags% -o bin\mpicxx.exe -DCXX %mpic% || popd && exit /b 1
copy bin\mpicxx.exe "bin\mpic++.exe"
gcc %cflags% -o bin\mpifort.exe -DFC %mpic% || popd && exit /b 1
copy bin\mpifort.exe bin\mpif77.exe
copy bin\mpifort.exe bin\mpif90.exe
rem Build FORTRAN90 MPI modules
bin\mpifort -c -Jinclude ..\mpi.f90 || popd && exit /b 1
del mpi.o
cd ..\..
rem Installation
xcopy /e/y bin %install_path:/=\%\%msmpi%\bin\
xcopy /e/y license %install_path:/=\%\%msmpi%\license\
xcopy /e/y sdk\build-x86_64 %install_path:/=\%\%msmpi%\
popd
exit /b 0

rem ********
rem OpenBLAS
rem ********
:build_openblas
call :download %openblas% "https://github.com/xianyi/OpenBLAS/archive/refs/tags/v0.3.21.tar.gz" || exit /b 1
set cmake_config_options=
call :cmake_build %openblas% || exit /b 1
exit /b 0

rem *********
rem ScaLAPACK
rem *********
:build_scalapack
call :download %scalapack% "https://github.com/Reference-ScaLAPACK/scalapack/archive/refs/tags/v2.1.0.zip" || exit /b 1
rem We need to use a foward slash for lapack and blas libraries paths since CMake CheckFortranFunctionExists is bugged when using a backward slash
set cmake_config_options=-DMPI_BASE_DIR:PATH=%install_path%/%msmpi% -DLAPACK_LIBRARIES:STRING=%install_path%/%openblas%/lib/libopenblas.a -DBLAS_LIBRARIES:STRING=%install_path%/%openblas%/lib/libopenblas.a -DBUILD_TESTING:BOOL=OFF
call :cmake_build %scalapack% || exit /b 1
exit /b 0

rem *****
rem MUMPS
rem *****
:build_mumps
call :download %mumps% "https://github.com/scivision/mumps/archive/refs/tags/v5.4.1.11.zip" || exit /b 1
set cmake_config_options=-DBLAS_LIBRARY=%install_path%/%openblas%/lib/libopenblas.a -DLAPACK_LIBRARY=%install_path%/%openblas%/lib/libopenblas.a -DSCALAPACK_ROOT:PATH=%install_path%/%scalapack%
if %compiler% == mingw (
  set cmake_config_options=%cmake_config_options% -DMPI_ROOT:PATH=%install_path%/%msmpi%
) else (
  if not defined MKLROOT (
    rem Set Intel MKL environment
    call "%ONEAPI_ROOT%mkl\latest\env\vars.bat"
  )
  if not defined I_MPI_ROOT (
    rem Set Intel MPI environment
    call "%I_MPI_ONEAPI_ROOT%\env\vars.bat"
  )
  set cmake_config_options=%cmake_config_options% -DMPI_ROOT:PATH="%I_MPI_ROOT%"
)
call :cmake_build %mumps% || exit /b 1
exit /b 0

rem ****
rem AED2
rem ****
:build_aed2
call :download %aed2% "http://www.opentelemac.org/index.php/component/jdownloads/finish/39-manual-installation-sources/2126-aed2m/0?Itemid=55" || exit /b 1
echo Building libaed2 with %compiler_name% compiler...
set install_dir=%install_path:/=\%\%aed2%
rd /s/q %install_dir% 2>nul
pushd %aed2%
mingw32-make || goto build_aed2_error
mkdir %install_dir%
xcopy /e/y include %install_dir%\include\
xcopy /y .\*.mod %install_dir%\include\
xcopy /y .\*.a %install_dir%\lib\
popd
exit /b 0

:build_aed2_error
popd
exit /b 1

rem ****
rem GOTM
rem ****
:build_gotm
call :download %gotm% "http://www.opentelemac.org/index.php/component/jdownloads/finish/39-manual-installation-sources/2127-gotm/0?Itemid=54" || exit /b 1
set cmake_config_options=-DBUILD_SHARED_LIBS:BOOL=ON -DGOTM_USE_NetCDF=OFF
call :cmake_build %gotm% || exit /b 1
exit /b 0

rem ***************
rem Python packages
rem ***************
:install_python_packages
rem Install all Python requirements for TELEMAC-MASCARET
echo Installing Python requirements...
rem Create a virtual environment
set python_dir=%install_path:/=\%\python
rd /s/q %python_dir% 2>nul
py -3.10 -m venv %python_dir% --prompt telemac
rem Copy python310.dll to the virtual environment (required by F2PY)
pushd %python_dir%
py -3.10 -c "import os, sys, shutil; shutil.copy2(f'{os.path.dirname(sys.executable)}\\python310.dll', f'{os.getcwd()}\\Scripts')"
rem Mingw also requires a definition file for F2PY in a "libs" subdirectory
if %compiler% == mingw (
  mkdir libs
  pushd libs
  py -3.10 -c "import os, sys; os.system(f'gendef {os.path.dirname(sys.executable)}\python310.dll')"
  popd
)
popd
rem Activate the virtual environment and install all Python requirements
call %python_dir%\Scripts\activate.bat
pushd %~dp0
pip install -r python_requirements.txt || goto install_error

rem Starting with Python 3.8, mpi4py requires MPI DLLs to be present in the module directory: copy them there
xcopy /y %install_path:/=\%\%msmpi%\bin\msmpi*.dll %python_dir%\Lib\site-packages\mpi4py\ >nul

popd
exit /b 0

:install_error
popd
exit /b 1

rem **************************************
rem Download a library if not already done
rem **************************************
:download
set lib=%~1
set url="%~2"

if exist %lib% (
  exit /b 0
)

rem Download and extract
echo Downloading and extracting %lib%...
set output=%lib%_arch
curl -L %url% -o %output% || exit /b 1
mkdir %lib%
tar -xvzf %output% --strip-components=1 -C %lib% || exit /b 1
del %output%
rem Apply patch if needed
if exist %~dp0patches\%lib%.patch (
  echo Patching %lib%...
  pushd %lib%
  git apply --ignore-space-change %~dp0patches\%lib%.patch || popd && exit /b 1
  popd
)

exit /b 0

rem ********************
rem CMake build function
rem ********************
:cmake_build
set lib=%~1

rem Set the number of build processes
set /a build_proc=%number_of_processors%/2

if %compiler% == mingw (
  set cmake_config_options=%cmake_config_options% -G "MinGW Makefiles" -DCMAKE_BUILD_TYPE:STRING=Release
  set cmake_build_options=-j %build_proc%
) else (
  if %compiler% == intel (
    set cmake_config_options=%cmake_config_options% -G "Visual Studio 16 2019" -A x64 -T "Intel C++ Compiler 19.2" -DCMAKE_CONFIGURATION_TYPES:STRING=Release
    set cmake_build_options=--config Release -j %build_proc%
  ) else (
    echo Unknown compiler: %compiler%.
    exit /b 1
  )
)

echo Building %lib% with %compiler_name% compiler...

pushd %lib%
mkdir build-%compiler% 2>nul
cd build-%compiler%
cmake %cmake_config_options% -DCMAKE_INSTALL_PREFIX:PATH=%install_path%/%lib% .. || goto cmake_build_error
cmake --build . %cmake_build_options% || goto cmake_build_error
cmake --install . || goto cmake_build_error
popd
color 7
exit /b 0

:cmake_build_error
popd
exit /b 1
